import pytest
from asynctest import Mock, patch, CoroutineMock
from tornado.httpclient import HTTPRequest, HTTPClientError, HTTPError
from tornado.httputil import HTTPServerRequest, HTTPHeaders

from botocore.credentials import Credentials

from aws_jupyter_proxy.awsproxy import AwsProxyRequest, create_endpoint_resolver


@pytest.fixture
def mock_session():
    session = Mock()
    session.get_credentials.return_value = Credentials(
        "access_key", "secret_key", "session_token"
    )
    return session


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_post_with_body(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="POST",
        uri="/awsproxy",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-west-2/sagemaker/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-target;x-amz-user-agent, "
                "Signature=cfe54b727d00698b9940531b1c9e456fd70258adc41fb338896455fddd6f3f2f",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "a83a35dcbd19cfad5b714cb12b5275a4cfa7e1012b633d9206300f09e058e7fa",
                "X-Amz-Target": "SageMaker.ListNotebookInstances",
                "X-Amz-Date": "20190816T204930Z",
            }
        ),
        body=b'{"NameContains":"myname"}',
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://api.sagemaker.us-west-2.amazonaws.com/",
        method=upstream_request.method,
        body=b'{"NameContains":"myname"}',
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190816/us-west-2/sagemaker/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-target;x-amz-user-agent, "
            "Signature="
            "215b2e3656147651194acb6cca20d5cb01dd8f396ac941533fc3e52b7cb563dc",
            "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
            "X-Amz-Content-Sha256": "a83a35dcbd19cfad5b714cb12b5275a4cfa7e1012b633d9206300f09e058e7fa",
            "X-Amz-Target": "SageMaker.ListNotebookInstances",
            "X-Amz-Date": "20190816T204930Z",
            "X-Amz-Security-Token": "session_token",
            "Host": "api.sagemaker.us-west-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_success_endpoint_override(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="POST",
        uri="/awsproxy",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-west-2/sagemaker/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-target;x-amz-user-agent, "
                "Signature=cfe54b727d00698b9940531b1c9e456fd70258adc41fb338896455fddd6f3f2f",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "a83a35dcbd19cfad5b714cb12b5275a4cfa7e1012b633d9206300f09e058e7fa",
                "X-Amz-Target": "SageMaker.ListNotebookInstances",
                "X-Amz-Date": "20190816T204930Z",
                "X-service-endpoint-url": "https://axis.us-west-2.amazonaws.com",
            }
        ),
        body=b'{"NameContains":"myname"}',
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://axis.us-west-2.amazonaws.com/",
        method=upstream_request.method,
        body=b'{"NameContains":"myname"}',
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190816/us-west-2/sagemaker/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-target;x-amz-user-agent, "
            "Signature="
            "e21ff53bc657eaa68ee08193cccdeb82027692d45d6ea57a5c3364e7fc95954e",
            "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
            "X-Amz-Content-Sha256": "a83a35dcbd19cfad5b714cb12b5275a4cfa7e1012b633d9206300f09e058e7fa",
            "X-Amz-Target": "SageMaker.ListNotebookInstances",
            "X-Amz-Date": "20190816T204930Z",
            "X-Service-Endpoint-Url": "https://axis.us-west-2.amazonaws.com",
            "Host": "axis.us-west-2.amazonaws.com",
            "X-Amz-Security-Token": "session_token",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_wrongurl_endpoint_override(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="POST",
        uri="/awsproxy",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-west-2/sagemaker/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-target;x-amz-user-agent, "
                "Signature=cfe54b727d00698b9940531b1c9e456fd70258adc41fb338896455fddd6f3f2f",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "a83a35dcbd19cfad5b714cb12b5275a4cfa7e1012b633d9206300f09e058e7fa",
                "X-Amz-Target": "SageMaker.ListNotebookInstances",
                "X-Amz-Date": "20190816T204930Z",
                "X-service-endpoint-url": "https://aws.amazon.com",
            }
        ),
        body=b'{"NameContains":"myname"}',
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://api.sagemaker.us-west-2.amazonaws.com/",
        method=upstream_request.method,
        body=b'{"NameContains":"myname"}',
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190816/us-west-2/sagemaker/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-target;x-amz-user-agent, "
            "Signature="
            "215b2e3656147651194acb6cca20d5cb01dd8f396ac941533fc3e52b7cb563dc",
            "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
            "X-Amz-Content-Sha256": "a83a35dcbd19cfad5b714cb12b5275a4cfa7e1012b633d9206300f09e058e7fa",
            "X-Amz-Target": "SageMaker.ListNotebookInstances",
            "X-Amz-Date": "20190816T204930Z",
            "X-Service-Endpoint-Url": "https://aws.amazon.com",
            "Host": "api.sagemaker.us-west-2.amazonaws.com",
            "X-Amz-Security-Token": "session_token",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_nonhttp_endpoint_override(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="POST",
        uri="/awsproxy",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-west-2/sagemaker/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-target;x-amz-user-agent, "
                "Signature=cfe54b727d00698b9940531b1c9e456fd70258adc41fb338896455fddd6f3f2f",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "a83a35dcbd19cfad5b714cb12b5275a4cfa7e1012b633d9206300f09e058e7fa",
                "X-Amz-Target": "SageMaker.ListNotebookInstances",
                "X-Amz-Date": "20190816T204930Z",
                "X-service-endpoint-url": "http://axis.us-west-2.amazonaws.com",
            }
        ),
        body=b'{"NameContains":"myname"}',
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://api.sagemaker.us-west-2.amazonaws.com/",
        method=upstream_request.method,
        body=b'{"NameContains":"myname"}',
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190816/us-west-2/sagemaker/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-target;x-amz-user-agent, "
            "Signature="
            "215b2e3656147651194acb6cca20d5cb01dd8f396ac941533fc3e52b7cb563dc",
            "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
            "X-Amz-Content-Sha256": "a83a35dcbd19cfad5b714cb12b5275a4cfa7e1012b633d9206300f09e058e7fa",
            "X-Amz-Target": "SageMaker.ListNotebookInstances",
            "X-Amz-Date": "20190816T204930Z",
            "X-Service-Endpoint-Url": "http://axis.us-west-2.amazonaws.com",
            "Host": "api.sagemaker.us-west-2.amazonaws.com",
            "X-Amz-Security-Token": "session_token",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_errors_passed_through(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="GET",
        uri="/awsproxy/clusters/myname",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-east-2/eks/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=f176ff82da6efc539bb8a2860be6ea19a99adf93d87be8ba96f25f1d29c91ba9",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190816T224244Z",
            }
        ),
        body=b"",
        host="localhost:8888",
    )
    mock_fetch.side_effect = HTTPClientError(code=500, message="Something bad")

    # When
    with pytest.raises(HTTPClientError) as e:
        await AwsProxyRequest(
            upstream_request, create_endpoint_resolver(), mock_session
        ).execute_downstream()

        assert 500 == e.value.code
        assert "Something bad" == e.value.message


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_get_with_path(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="GET",
        uri="/awsproxy/clusters/myname",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-east-2/eks/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=f176ff82da6efc539bb8a2860be6ea19a99adf93d87be8ba96f25f1d29c91ba9",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190816T224244Z",
            }
        ),
        body=b"",
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://eks.us-east-2.amazonaws.com/clusters/myname",
        method=upstream_request.method,
        body=None,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190816/us-east-2/eks/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "0bab991ebb02f7f2ea44e3687778e930e5f66e1111b958a9c2ff88aba2eaf3da",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "eks.us-east-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_get_with_query(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="GET",
        uri="/awsproxy/clusters?maxResults=1",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-east-2/eks/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=24203e07130d74b28845f756f5440603d24400d53d07ddda9d7add99d5ec7c8d",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190816T224244Z",
            }
        ),
        body=b"",
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://eks.us-east-2.amazonaws.com/clusters?maxResults=1",
        method=upstream_request.method,
        body=None,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190816/us-east-2/eks/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "61315054f93efa316230cbe77497522b8db692969104ec4c935235e14ad1c23f",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "eks.us-east-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_delete(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="DELETE",
        uri="/awsproxy/clusters/myname",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-east-2/eks/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=24203e07130d74b28845f756f5440603d24400d53d07ddda9d7add99d5ec7c8d",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190816T224244Z",
            }
        ),
        body=b"",
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://eks.us-east-2.amazonaws.com/clusters/myname",
        method=upstream_request.method,
        body=None,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190816/us-east-2/eks/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "6fde2ff4f8aa9582d8740b6b7108a1a7f24ec3807a15c79f6e688ef4f4eaae35",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "eks.us-east-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_put(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="PUT",
        uri="/awsproxy/clusters/myname",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-east-2/eks/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=24203e07130d74b28845f756f5440603d24400d53d07ddda9d7add99d5ec7c8d",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190816T224244Z",
            }
        ),
        body=b'{"Name":"Foo","Id":"Bar"}',
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://eks.us-east-2.amazonaws.com/clusters/myname",
        method=upstream_request.method,
        body=upstream_request.body,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190816/us-east-2/eks/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "b6d04b91fa9e1993657806821321eb10a665e89d6de2f390fa39d40d77015971",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "eks.us-east-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_post_with_query_params_and_body(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="POST",
        uri="/awsproxy/bucket-name-1/Hello.txt?select&select-type=2",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190828/us-west-2/s3/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=451d7037903cee381c6a5d2c61c6ee2b5d36f35650e95abcf5e8af11b57c0cf8",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "f5b0fecc75eeaba2a9c92703363f4b9e3efa3f7811d3904f0b71cf05c3895228",
                "X-Amz-Date": "20190828T173626Z",
            }
        ),
        body=b'<SelectObjectContentRequest xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
        b"<Expression>select * from S3Object</Expression>"
        b"<ExpressionType>SQL</ExpressionType>"
        b"<InputSerialization><JSON>"
        b"<Type>Lines</Type>"
        b"</JSON></InputSerialization>"
        b"<OutputSerialization><JSON><RecordDelimiter>,</RecordDelimiter></JSON></OutputSerialization>"
        b"</SelectObjectContentRequest>",
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://s3.us-west-2.amazonaws.com/bucket-name-1/Hello.txt?select&select-type=2",
        method=upstream_request.method,
        body=upstream_request.body,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190828/us-west-2/s3/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "e55d97f45ca6862d9c2518868dd2a8c383007df1c991e42ef5950a46a4c13f8e",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "s3.us-west-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_post_with_query_params_no_body(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="POST",
        uri="/awsproxy/bucket-name-1/Multipart-0.16441670919496487.txt?uploads",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190828/us-west-2/s3/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=451d7037903cee381c6a5d2c61c6ee2b5d36f35650e95abcf5e8af11b57c0cf8",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "a08b81ef3b4ec5e1f65ca97d00928105c3d7eb6d50ae59fc15f0d14b64c9ec3b",
                "X-Amz-Date": "20190828T173626Z",
            }
        ),
        body=b"",
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://s3.us-west-2.amazonaws.com/bucket-name-1/Multipart-0.16441670919496487.txt"
        "?uploads",
        method=upstream_request.method,
        body=None,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190828/us-west-2/s3/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "ba2062818cb4cd80a73dd43d006f141ede069b8ccc2ece16c20504587bd5045b",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "s3.us-west-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_head_request(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="HEAD",
        uri="/awsproxy/bucket-name-1",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190828/us-west-2/s3/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=0d02795c4feed38e5a4cd80aec3a2c67886b11797a23c307e4f52c2cfe0c137e",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190828T173626Z",
            }
        ),
        body=None,
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://s3.us-west-2.amazonaws.com/bucket-name-1",
        method=upstream_request.method,
        body=None,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190828/us-west-2/s3/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "6d724e3bd64390d5d84010d6fc0f8147b3e3917c5befa3f8d1efb691b408e821",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "s3.us-west-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
async def test_get_with_encoded_uri(mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="GET",
        uri="/awsproxy/bucket-name-1/ll%3A%3Askeleton%201.png",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190816/us-west-2/s3/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=822f116d22d577aa2dc1033f354fa2a6fd3a2b6a0fd51885472b57daf45d605e",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190816T224244Z",
            }
        ),
        body=b"",
        host="localhost:8888",
    )

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://s3.us-west-2.amazonaws.com/bucket-name-1/ll%3A%3Askeleton%201.png",
        method=upstream_request.method,
        body=None,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190816/us-west-2/s3/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "4715991ba2461bfda29bda8a53747a13448c1303c2e03d8ee8a4992df08f5551",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "s3.us-west-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("os.getenv")
async def test_request_not_whitelisted(mock_getenv, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="HEAD",
        uri="/awsproxy/bucket-name-1",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190828/us-west-2/s3/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=0d02795c4feed38e5a4cd80aec3a2c67886b11797a23c307e4f52c2cfe0c137e",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190828T173626Z",
            }
        ),
        body=None,
        host="localhost:8888",
    )
    mock_getenv.return_value = "sagemaker,eks,"

    # When
    with pytest.raises(HTTPError) as e:
        await AwsProxyRequest(
            upstream_request, create_endpoint_resolver(), mock_session
        ).execute_downstream()

        # Then
        assert 403 == e.value.code
        assert "Service s3 is not whitelisted for proxying requests" == e.value.message


@pytest.mark.asyncio
@patch("os.getenv")
async def test_nothing_whitelisted(mock_getenv, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="HEAD",
        uri="/awsproxy/bucket-name-1",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190828/us-west-2/s3/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=0d02795c4feed38e5a4cd80aec3a2c67886b11797a23c307e4f52c2cfe0c137e",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190828T173626Z",
            }
        ),
        body=None,
        host="localhost:8888",
    )
    mock_getenv.return_value = ""

    # When
    with pytest.raises(HTTPError) as e:
        await AwsProxyRequest(
            upstream_request, create_endpoint_resolver(), mock_session
        ).execute_downstream()

        # Then
        assert 403 == e.value.code
        assert "Service s3 is not whitelisted for proxying requests" == e.value.message


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
@patch("os.getenv")
async def test_request_whitelisted(mock_getenv, mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="HEAD",
        uri="/awsproxy/bucket-name-1",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190828/us-west-2/s3/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=0d02795c4feed38e5a4cd80aec3a2c67886b11797a23c307e4f52c2cfe0c137e",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190828T173626Z",
            }
        ),
        body=None,
        host="localhost:8888",
    )
    mock_getenv.return_value = "s3,"

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://s3.us-west-2.amazonaws.com/bucket-name-1",
        method=upstream_request.method,
        body=None,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190828/us-west-2/s3/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "6d724e3bd64390d5d84010d6fc0f8147b3e3917c5befa3f8d1efb691b408e821",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "s3.us-west-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


@pytest.mark.asyncio
@patch("tornado.httpclient.AsyncHTTPClient.fetch", new_callable=CoroutineMock)
@patch("os.getenv")
async def test_request_with_base_url(mock_getenv, mock_fetch, mock_session):
    # Given
    upstream_request = HTTPServerRequest(
        method="HEAD",
        uri="base-url/awsproxy/bucket-name-1",
        headers=HTTPHeaders(
            {
                "Authorization": "AWS4-HMAC-SHA256 "
                "Credential=AKIDEXAMPLE/20190828/us-west-2/s3/aws4_request, "
                "SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-user-agent, "
                "Signature=0d02795c4feed38e5a4cd80aec3a2c67886b11797a23c307e4f52c2cfe0c137e",
                "Host": "localhost:8888",
                "X-Amz-User-Agent": "aws-sdk-js/2.507.0 promise",
                "X-Amz-Content-Sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                "X-Amz-Date": "20190828T173626Z",
            }
        ),
        body=None,
        host="localhost:8888",
    )
    mock_getenv.return_value = "s3,"

    # When
    await AwsProxyRequest(
        upstream_request, create_endpoint_resolver(), mock_session
    ).execute_downstream()

    # Then
    expected = HTTPRequest(
        url="https://s3.us-west-2.amazonaws.com/bucket-name-1",
        method=upstream_request.method,
        body=None,
        headers={
            "Authorization": "AWS4-HMAC-SHA256 "
            "Credential=access_key/20190828/us-west-2/s3/aws4_request, "
            "SignedHeaders=host;x-amz-content-sha256;x-amz-date;"
            "x-amz-security-token;x-amz-user-agent, "
            "Signature="
            "6d724e3bd64390d5d84010d6fc0f8147b3e3917c5befa3f8d1efb691b408e821",
            "X-Amz-User-Agent": upstream_request.headers["X-Amz-User-Agent"],
            "X-Amz-Content-Sha256": upstream_request.headers["X-Amz-Content-Sha256"],
            "X-Amz-Date": upstream_request.headers["X-Amz-Date"],
            "X-Amz-Security-Token": "session_token",
            "Host": "s3.us-west-2.amazonaws.com",
        },
        follow_redirects=False,
        allow_nonstandard_methods=True,
    )

    assert_http_response(mock_fetch, expected)


def assert_http_response(mock_fetch, expected_http_request):
    mock_fetch.assert_awaited_once()
    actual_http_request: HTTPRequest = mock_fetch.await_args[0][0]
    assert expected_http_request.url == actual_http_request.url
    assert expected_http_request.body == actual_http_request.body
    assert expected_http_request.method == actual_http_request.method
    assert (
        expected_http_request.follow_redirects == actual_http_request.follow_redirects
    )
    assert (
        expected_http_request.allow_nonstandard_methods
        == actual_http_request.allow_nonstandard_methods
    )
    assert expected_http_request.headers == actual_http_request.headers



import hashlib
import hmac
import os
import re
from collections import namedtuple
from functools import lru_cache
from typing import List, Tuple
from urllib.parse import urlparse, urlunparse, quote

from botocore.client import ClientEndpointBridge
from botocore.loaders import create_loader
from botocore.model import ServiceModel
from botocore.regions import EndpointResolver
from botocore.session import Session
from notebook.base.handlers import APIHandler
from tornado.httpclient import (
    AsyncHTTPClient,
    HTTPRequest,
    HTTPResponse,
    HTTPClientError,
    HTTPError,
)
from tornado.httputil import HTTPServerRequest, HTTPHeaders

ServiceInfo = namedtuple(
    "ServiceInfo", ["service_name", "host", "endpoint_url", "credential_scope"]
)
UpstreamAuthInfo = namedtuple(
    "UpstreamAuthInfo", ["service_name", "region", "signed_headers"]
)


# maxsize is arbitrarily taken from https://docs.python.org/3/library/functools.html#functools.lru_cache
@lru_cache(maxsize=128)
def get_service_info(
    endpoint_resolver: EndpointResolver,
    service_name: str,
    region: str,
    endpoint_override: str,
) -> ServiceInfo:
    service_model_json = create_loader().load_service_model(service_name, "service-2")

    service_data = ClientEndpointBridge(endpoint_resolver).resolve(
        service_name=ServiceModel(
            service_model_json, service_name=service_name
        ).endpoint_prefix,
        region_name=region,
    )

    if endpoint_override and re.fullmatch(
        r"https:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.(aws.dev|amazonaws.com)\b",
        endpoint_override,
    ):
        service_data["endpoint_url"] = endpoint_override

    return ServiceInfo(
        service_name,
        service_data["metadata"]["hostname"],
        service_data["endpoint_url"],
        service_data["metadata"].get("credentialScope"),
    )


def create_endpoint_resolver() -> EndpointResolver:
    """
    Creates an instance of the botocore EndpointResolver. Used to inject the instance during application initialization
    to avoid loading endpoint data on a per-request basis.
    :return: the EndpointResolver instance
    """
    return EndpointResolver(create_loader().load_data("endpoints"))


class AwsProxyHandler(APIHandler):
    def initialize(self, endpoint_resolver: EndpointResolver, session: Session):
        """
        Hook for Tornado handler initialization.
        :param session: the botocore session
        :param endpoint_resolver: the application level EndpointResolver instance
        """
        self.endpoint_resolver = endpoint_resolver
        self.session = session

    async def handle_request(self):
        try:
            response = await AwsProxyRequest(
                self.request, self.endpoint_resolver, self.session
            ).execute_downstream()
            self.set_status(response.code, response.reason)
            self._finish_response(response)
        except HTTPClientError as e:
            self.set_status(e.code, e.message)
            if e.response:
                self._finish_response(e.response)
            else:
                super(APIHandler, self).finish()

    def _finish_response(self, response: HTTPResponse):
        for name, value in response.headers.get_all():
            if self._is_blacklisted_response_header(name, value):
                continue
            self.set_header(name, value)
        csp_value = response.headers.get(
            "Content-Security-Policy", "upgrade-insecure-requests; base-uri 'none';"
        )
        self.set_header("Content-Security-Policy", csp_value)
        super(APIHandler, self).finish(response.body or None)

    async def post(self, *args):
        await self.handle_request()

    async def get(self, *args):
        await self.handle_request()

    async def delete(self, *args):
        await self.handle_request()

    async def patch(self, *args):
        await self.handle_request()

    async def put(self, *args):
        await self.handle_request()

    async def head(self, *args):
        await self.handle_request()

    @staticmethod
    def _is_blacklisted_response_header(name: str, value: str) -> bool:
        if name == "Transfer-Encoding" and value == "chunked":
            # Responses are no longer "chunked" when we send them to the browser.
            # If we retain this header, then the browser will wait forever for more chunks.
            return True
        elif name == "Content-Length":
            # Tornado will auto-set the Content-Length
            return True
        else:
            return False


class AwsProxyRequest(object):
    """
    A class representing a request being proxied from an upstream client (browser) to the downstream AWS service.
    """

    BLACKLISTED_REQUEST_HEADERS: List[str] = ["Origin", "Host"]

    def __init__(
        self,
        upstream_request: HTTPServerRequest,
        endpoint_resolver: EndpointResolver,
        session: Session,
    ):
        """
        :param upstream_request: The original upstream HTTP request from the client(browser) to Jupyter
        :param endpoint_resolver: The botocore endpoint_resolver instance
        """
        self.upstream_request = upstream_request
        self.endpoint_resolver = endpoint_resolver

        self.credentials = session.get_credentials()

        self.upstream_auth_info = self._build_upstream_auth_info()
        self.service_info = get_service_info(
            endpoint_resolver,
            self.upstream_auth_info.service_name,
            self.upstream_auth_info.region,
            self.upstream_request.headers.get("X-service-endpoint-url", None),
        )
        # if the environment variable is not specified, os.getenv returns None, and no whitelist is in effect.
        self.whitelisted_services = (
            os.getenv("AWS_JUPYTER_PROXY_WHITELISTED_SERVICES").strip(",").split(",")
            if os.getenv("AWS_JUPYTER_PROXY_WHITELISTED_SERVICES") is not None
            else None
        )

    async def execute_downstream(self) -> HTTPResponse:
        """
        Executes the downstream request (Jupyter to AWS service) and return the response or the error
        after adding SigV4 authentication.

        "allow_nonstandard_methods" is used because Tornado rejects POST requests without a body without this parameter,
        and some operations send such requests (such as S3.InitiateMultipartUpload)
        :return: the HTTPResponse
        """
        if (
            self.whitelisted_services is not None
            and self.service_info.service_name not in self.whitelisted_services
        ):
            raise HTTPError(
                403,
                message=f"Service {self.service_info.service_name} is not whitelisted for proxying requests",
            )

        base_service_url = urlparse(self.service_info.endpoint_url)
        start_index = self.upstream_request.path.index("/awsproxy") + len("/awsproxy")
        downstream_request_path = (
            base_service_url.path + self.upstream_request.path[start_index:] or "/"
        )
        return await AsyncHTTPClient().fetch(
            HTTPRequest(
                method=self.upstream_request.method,
                url=self._compute_downstream_url(downstream_request_path),
                headers=self._compute_downstream_headers(downstream_request_path),
                body=self.upstream_request.body or None,
                follow_redirects=False,
                allow_nonstandard_methods=True,
            )
        )

    def _compute_downstream_url(self, downstream_request_path) -> str:
        base_service_url = urlparse(self.service_info.endpoint_url)
        return urlunparse(
            [
                base_service_url.scheme,
                base_service_url.netloc,
                downstream_request_path,
                base_service_url.params,
                self.upstream_request.query,
                None,
            ]
        )

    def _compute_downstream_headers(self, downstream_request_path) -> HTTPHeaders:
        """
        1. Copy original headers apart from blacklisted ones
        2. Add the Host header based on the service model
        3. Add a security token header if the current session is using temporary credentials
        4. Add the SigV4 Authorization header.

        :param downstream_request_path: the URL path for the downstream service request
        :return: the headers to pass to the downstream request
        """
        downstream_request_headers = self.upstream_request.headers.copy()
        for blacklisted_request_header in self.BLACKLISTED_REQUEST_HEADERS:
            try:
                del downstream_request_headers[blacklisted_request_header]
            except KeyError:
                pass

        base_service_url = urlparse(self.service_info.endpoint_url)
        downstream_request_headers["Host"] = base_service_url.netloc

        if self.credentials.token:
            downstream_request_headers["X-Amz-Security-Token"] = self.credentials.token

        downstream_request_headers["Authorization"] = self._sigv4_auth_header(
            downstream_request_path
        )
        return downstream_request_headers

    def _sigv4_auth_header(self, downstream_request_path) -> str:
        """
        Computes the SigV4 signature following https://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html

        :param downstream_request_path: the URL path for the downstream service's request
        :return: the Authorization header containing SigV4 credetntials
        """
        # ************* TASK 1: CREATE THE CANONICAL REQUEST*************
        canonical_method = self.upstream_request.method
        canonical_uri = quote(downstream_request_path)
        canonical_querystring = self._get_canonical_querystring()
        signed_headers, canonical_headers = self._get_signed_canonical_headers()
        payload_hash = hashlib.sha256(self.upstream_request.body).hexdigest()

        canonical_request = (
            f"{canonical_method}\n"
            f"{canonical_uri}\n"
            f"{canonical_querystring}\n"
            f"{canonical_headers}\n"
            f"{signed_headers}\n"
            f"{payload_hash}"
        )

        # ************* TASK 2: CREATE THE STRING TO SIGN*************
        algorithm = "AWS4-HMAC-SHA256"
        region = self._get_downstream_signing_region()
        amz_date = self.upstream_request.headers["X-Amz-Date"]
        date_stamp = amz_date[0:8]

        credential_scope = (
            f"{date_stamp}/{region}/{self.service_info.service_name}/aws4_request"
        )
        request_digest = hashlib.sha256(canonical_request.encode("utf-8")).hexdigest()
        string_to_sign = (
            f"{algorithm}\n" f"{amz_date}\n" f"{credential_scope}\n" f"{request_digest}"
        )

        # ************* TASK 3: CALCULATE THE SIGNATURE *************
        signing_key = get_signature_key(
            self.credentials.secret_key,
            date_stamp,
            region,
            self.service_info.service_name,
        )
        signature = hmac.new(
            signing_key, string_to_sign.encode("utf-8"), hashlib.sha256
        ).hexdigest()

        # ************* TASK 4: BUILD THE AUTH HEADER *************
        authorization_header = (
            f"{algorithm} "
            f"Credential={self.credentials.access_key}/{credential_scope}, "
            f"SignedHeaders={signed_headers}, "
            f"Signature={signature}"
        )

        return authorization_header

    def _get_canonical_querystring(self) -> str:
        canonical_query_string = ""
        corrected_request_query = self.upstream_request.query.replace("+", "%20")
        if corrected_request_query != "":
            query_string_list = []
            for item in corrected_request_query.split("&"):
                query_string_part = item.split("=", maxsplit=1)
                if len(query_string_part) == 2:
                    query_string_list.append(query_string_part)
                elif len(query_string_part) == 1:
                    query_string_part.append("")
                    query_string_list.append(query_string_part)
                else:
                    raise ValueError(f"Invalid query string split for {item}")
            query_string_dict = dict(query_string_list)
            sorted_q_string_list = [
                f"{k}={query_string_dict[k]}" for k in sorted(query_string_dict)
            ]
            canonical_query_string = "&".join(sorted_q_string_list)
        return canonical_query_string

    def _get_signed_canonical_headers(self) -> Tuple[str, str]:
        canonical_headers = {}

        for signed_header in self.upstream_auth_info.signed_headers:
            canonical_headers[signed_header] = self.upstream_request.headers[
                signed_header
            ]

        base_service_url = urlparse(self.service_info.endpoint_url)
        canonical_headers["host"] = base_service_url.netloc
        if self.credentials.token:
            canonical_headers["x-amz-security-token"] = self.credentials.token

        canonical_headers_string = "\n".join(
            [
                f"{canonical_header}:{canonical_headers[canonical_header]}"
                for canonical_header in sorted(canonical_headers)
            ]
        )
        canonical_headers_string += "\n"
        signed_headers = ";".join(sorted(canonical_headers))

        return signed_headers, canonical_headers_string

    def _get_downstream_signing_region(self) -> str:
        """
        Get the region to sign the downstream request for. The default is the region that the request was originally
        signed,  but if the service has a credentialScope override specified in the service config then that is used.
        :return: the region to sign the request with.
        """
        if not self.service_info.credential_scope:
            return self.upstream_auth_info.region

        try:
            return self.service_info.credential_scope["region"]
        except KeyError:
            return self.upstream_auth_info.region

    def _build_upstream_auth_info(self) -> UpstreamAuthInfo:
        """
        Parses the upstream requests's Authorization header to determine identifying information such as the region and
        the service the request was originally signed for.

        Sample header:
            AWS4-HMAC-SHA256 \
            Credential=SOMEACCESSKEY/20190814/aws_region/aws_service/aws4_request, \
            SignedHeaders=host;x-amz-content-sha256;x-amz-date;x-amz-target;x-amz-user-agent, \
            Signature=e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855

        :return: the UpstreamAuthInfo instance
        """
        auth_header_parts = self.upstream_request.headers["Authorization"].split(" ")

        signed_headers = auth_header_parts[2].strip(",").split("=")[1].split(";")
        _, _, region, service_name, _ = auth_header_parts[1].split("=")[1].split("/")
        return UpstreamAuthInfo(service_name, region, signed_headers)


# Key derivation functions. See:
# http://docs.aws.amazon.com/general/latest/gr/signature-v4-examples.html#signature-v4-examples-python
def sign(key, msg):
    return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()


def get_signature_key(key, date_stamp, region_name, service_name):
    k_date = sign(("AWS4" + key).encode("utf-8"), date_stamp)
    k_region = sign(k_date, region_name)
    k_service = sign(k_region, service_name)
    k_signing = sign(k_service, "aws4_request")
    return k_signing



from unittest.mock import patch
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application
from aws_jupyter_proxy.awsconfig import AwsConfigHandler
from aws_jupyter_proxy.handlers import awsproxy_handlers


class TestHelloApp(AsyncHTTPTestCase):
    def get_app(self):
        return Application(awsproxy_handlers)

    @patch("aws_jupyter_proxy.awsconfig.get_session")
    def test_valid_region(self, mock_get_session):
        mock_get_session.return_value.get_config_variable.return_value = "us-west-1"
        response = self.fetch("/awsproxy/awsconfig")
        self.assertEqual(response.code, 200)
        self.assertEqual(response.body, b'{"region": "us-west-1"}')

    @patch("aws_jupyter_proxy.awsconfig.get_session")
    def test_default_region(self, mock_get_session):
        mock_get_session.return_value.get_config_variable.return_value = ""
        response = self.fetch("/awsproxy/awsconfig")
        self.assertEqual(response.code, 200)
        self.assertEqual(response.body, b'{"region": null}')

    @patch("aws_jupyter_proxy.awsconfig.get_session")
    def test_no_region(self, mock_get_session):
        mock_get_session.return_value.get_config_variable.return_value = None
        response = self.fetch("/awsproxy/awsconfig")
        self.assertEqual(response.code, 200)
        self.assertEqual(response.body, b'{"region": null}')



from io import BytesIO

import tornado.web
from asynctest import CoroutineMock
from unittest.mock import patch
from tornado.httpclient import HTTPClientError, HTTPResponse, HTTPRequest
from tornado.httputil import HTTPHeaders
from tornado.testing import AsyncHTTPTestCase

from aws_jupyter_proxy.handlers import awsproxy_handlers


class TestAwsProxyHandler(AsyncHTTPTestCase):
    @patch("aws_jupyter_proxy.awsproxy.AwsProxyRequest")
    def test_downstream_error_no_body(self, mock_awsproxy):
        # Given
        mock_instance = mock_awsproxy.return_value
        mock_instance.execute_downstream = CoroutineMock()
        mock_instance.execute_downstream.side_effect = HTTPClientError(code=500)

        # When
        response = self.fetch("/awsproxy")

        # Then
        assert 500 == response.code
        assert b"" == response.body

    @patch("aws_jupyter_proxy.awsproxy.AwsProxyRequest")
    def test_downstream_error_with_body(self, mock_awsproxy):
        # Given
        mock_execute_downstream = CoroutineMock()
        mock_execute_downstream.side_effect = HTTPClientError(
            code=403,
            response=HTTPResponse(
                request=HTTPRequest("/foo"), code=403, buffer=BytesIO(b"AccessDenied")
            ),
        )

        mock_instance = mock_awsproxy.return_value
        mock_instance.execute_downstream = mock_execute_downstream

        # When
        response = self.fetch("/awsproxy")

        # Then
        mock_execute_downstream.assert_awaited_once()
        assert 403 == response.code
        assert b"AccessDenied" == response.body

    @patch("aws_jupyter_proxy.awsproxy.AwsProxyRequest")
    def test_downstream_success_blacklisted_headers_removed(self, mock_awsproxy):
        # Given
        mock_execute_downstream = CoroutineMock()
        mock_execute_downstream.return_value = HTTPResponse(
            request=HTTPRequest(url="https://awsservice.amazonaws.com/"),
            code=200,
            headers=HTTPHeaders(
                {
                    "Host": "awsservice.amazonaws.com",
                    "X-Amz-RequestId": "foo-abc",
                    "Transfer-Encoding": "chunked",
                }
            ),
            buffer=BytesIO(b"SomeResponse"),
        )

        mock_instance = mock_awsproxy.return_value
        mock_instance.execute_downstream = mock_execute_downstream

        # When
        response = self.fetch("/awsproxy")

        # Then
        mock_execute_downstream.assert_awaited_once()
        assert 200 == response.code
        assert b"SomeResponse" == response.body
        assert "Transfer-Encoding" not in response.headers
        assert "foo-abc" == response.headers["X-Amz-RequestId"]
        assert "awsservice.amazonaws.com" == response.headers["Host"]

    @patch("aws_jupyter_proxy.awsproxy.AwsProxyRequest")
    def test_downstream_success_with_content_security_policy(self, mock_awsproxy):
        # Given
        mock_execute_downstream = CoroutineMock()
        mock_execute_downstream.return_value = HTTPResponse(
            request=HTTPRequest(url="https://awsservice.amazonaws.com/"),
            code=200,
            headers=HTTPHeaders(
                {
                    "Host": "awsservice.amazonaws.com",
                    "X-Amz-RequestId": "foo-abc",
                    "Transfer-Encoding": "chunked",
                    "Content-Security-Policy": "default-src 'none';",
                }
            ),
            buffer=BytesIO(b"SomeResponse"),
        )

        mock_instance = mock_awsproxy.return_value
        mock_instance.execute_downstream = mock_execute_downstream

        # When
        response = self.fetch("/awsproxy")

        # Then
        mock_execute_downstream.assert_awaited_once()
        assert 200 == response.code
        assert b"SomeResponse" == response.body
        assert "Transfer-Encoding" not in response.headers
        assert "Content-Security-Policy" in response.headers
        assert "default-src 'none';" == response.headers["Content-Security-Policy"]
        assert "foo-abc" == response.headers["X-Amz-RequestId"]
        assert "awsservice.amazonaws.com" == response.headers["Host"]

    @patch("aws_jupyter_proxy.awsproxy.AwsProxyRequest")
    def test_downstream_success_without_content_security_policy(self, mock_awsproxy):
        # Given
        mock_execute_downstream = CoroutineMock()
        mock_execute_downstream.return_value = HTTPResponse(
            request=HTTPRequest(url="https://awsservice.amazonaws.com/"),
            code=200,
            headers=HTTPHeaders(
                {
                    "Host": "awsservice.amazonaws.com",
                    "X-Amz-RequestId": "foo-abc",
                    "Transfer-Encoding": "chunked",
                }
            ),
            buffer=BytesIO(b"SomeResponse"),
        )

        mock_instance = mock_awsproxy.return_value
        mock_instance.execute_downstream = mock_execute_downstream

        # When
        response = self.fetch("/awsproxy")

        # Then
        mock_execute_downstream.assert_awaited_once()
        assert 200 == response.code
        assert b"SomeResponse" == response.body
        assert "Transfer-Encoding" not in response.headers
        assert "Content-Security-Policy" in response.headers
        assert (
            "upgrade-insecure-requests; base-uri 'none';"
            == response.headers["Content-Security-Policy"]
        )
        assert "foo-abc" == response.headers["X-Amz-RequestId"]
        assert "awsservice.amazonaws.com" == response.headers["Host"]

    def get_app(self):
        return tornado.web.Application(awsproxy_handlers)



import json
from botocore.session import get_session
from notebook.base.handlers import APIHandler


class AwsConfigHandler(APIHandler):
    async def get(self, *args):
        response = {"region": self._get_aws_region()}
        self.write(json.dumps(response))

    def _get_aws_region(self):
        session = get_session()
        return session.get_config_variable("region") or None
